@@ -24,28 +24,32 @@
 #include <string>
 #include <stdexcept>
 #include <algorithm>
+#include "SharedData.hpp"
 
 #define LIM(g , min, max) ((g) > (max) ? (max) : ((g) < (min) ? (min) : (g)))
 
+static SharedData sharedData[4] = {SharedData(), SharedData(), SharedData(), SharedData()};
+
 BChoppr::BChoppr (double samplerate, const LV2_Feature* const* features) :
 	map(NULL),
 	rate(samplerate), bpm(120.0f), speed(1), position(0),
 	beatsPerBar (4), beatUnit (4), refFrame(0),
 	prevStep(0), actStep(0), nextStep(1),
 	audioInput1(NULL), audioInput2(NULL), audioOutput1(NULL), audioOutput2(NULL),
-	controllers {nullptr},
-	sequencesperbar (4), swing (1.0), nrSteps(16),
-	bypass (false), drywet (1.0f), blend (1), attack(0.2), release (0.2),
+	sharedDataNr (0),
+	controlPtrs {nullptr}, controllers {0.0f},
 	stepPositions {0.0},
 	controlPort1(NULL), controlPort2(NULL),  notifyPort(NULL),
-	record_on(false), monitorpos(-1), message ()
+	record_on(false),
+	notify_sharedData (false),
+	notify_controllers {false},
+	monitorpos(-1), message ()
 
 {
 	// Init array members
 	notifications.fill (defaultNotification);
 	monitor.fill (defaultMonitorData);
 	std::fill (stepAutoPositions, stepAutoPositions + MAXSTEPS - 1, true);
-	std::fill (stepLevels, stepLevels + MAXSTEPS, 1.0f);
 
 	//Scan host features for URID map
 	LV2_URID_Map* m = NULL;
@@ -68,7 +72,10 @@ BChoppr::BChoppr (double samplerate, const LV2_Feature* const* features) :
 	recalculateAutoPositions ();
 }
 
-BChoppr::~BChoppr () {}
+BChoppr::~BChoppr ()
+{
+	if ((sharedDataNr > 0) && (sharedDataNr <= 4)) sharedData[sharedDataNr - 1].unlink (this);
+}
 
 void BChoppr::connect_port(uint32_t port, void *data)
 {
@@ -95,7 +102,7 @@ void BChoppr::connect_port(uint32_t port, void *data)
 		audioOutput2 = (float*) data;
 		break;
 	default:
-		if ((port >= Controllers) && (port < Controllers + NrControllers)) controllers[port - Controllers] = (float*) data;
+		if ((port >= Controllers) && (port < Controllers + NrControllers)) controlPtrs[port - Controllers] = (float*) data;
 	}
 }
 
@@ -106,51 +113,19 @@ void BChoppr::run (uint32_t n_samples)
 	if ((!controlPort1) || (!controlPort2) || (!notifyPort) || (!audioInput1) || (!audioInput2) || (!audioOutput1) || (!audioOutput2)) return;
 	for (int i = 0; i < NrControllers; ++i)
 	{
-		if (!controllers[i]) return;
+		if (!controlPtrs[i]) return;
 	}
 
 	// Update controller values
-	bypass = (*controllers[Bypass - Controllers] != 0.0f);
-	drywet = LIM (*(controllers[DryWet - Controllers]), 0.0f, 1.0f);
-	blend = LIM (*(controllers[Blend - Controllers]), 1, 2);
-	attack = LIM (*(controllers[Attack - Controllers]), 0.01, 1.0);
-	release = LIM (*(controllers[Release - Controllers]), 0.01, 1.0);
-	sequencesperbar = LIM (round (*(controllers[SequencesPerBar - Controllers])), 1, 8);
-
-	float new_swing = LIM (*(controllers[Swing - Controllers]), 0.333333, 3.0);
-	if (new_swing != swing)
+	for (int i = 0; i < NrControllers; ++i)
 	{
-		swing = new_swing;
-		recalculateAutoPositions ();
-	}
 
-	float new_nrSteps = LIM (round(*(controllers[NrSteps - Controllers])), 1, 16);
-	if (new_nrSteps != nrSteps)
-	{
-		nrSteps = new_nrSteps;
-		recalculateAutoPositions ();
-	}
-
-	for (int i = 0; i < MAXSTEPS; ++i) stepLevels[i] = LIM (*(controllers[i + StepLevels - Controllers]), 0.0, 1.0);
-	for (int i = 0; i < MAXSTEPS - 1; ++i)
-	{
-		 float value = LIM (*(controllers[i + StepPositions - Controllers]), 0.0, 1.0);
-
-		 if (value == 0.0f)
-		 {
-			 if (!stepAutoPositions[i])
-			 {
-				 stepAutoPositions[i] = true;
-				 recalculateAutoPositions();
-			 }
-		 }
-
-		 else if (stepPositions[i] != value)
-		 {
-			 stepAutoPositions[i] = false;
-			 stepPositions[i] = value;
-			 recalculateAutoPositions();
-		 }
+		float newValue = readController (i);
+		if (newValue != controllers[i])
+		{
+			setController (i, newValue);
+			if ((sharedDataNr >= 1) && (sharedDataNr <= 4)) notify_controllers[i] = true;
+		}
 	}
 
 	// Process GUI data
@@ -160,8 +135,85 @@ void BChoppr::run (uint32_t n_samples)
 		if (lv2_atom_forge_is_object_type(&forge, ev2->body.type))
 		{
 			const LV2_Atom_Object* obj2 = (const LV2_Atom_Object*)&ev2->body;
-			if (obj2->body.otype == uris.ui_on) record_on = true;
+
+			// UI on
+			if (obj2->body.otype == uris.ui_on)
+			{
+				record_on = true;
+				notify_sharedData = true;
+				std::fill (notify_controllers, notify_controllers + NrControllers, true);
+			}
+
+			// UI off
 			else if (obj2->body.otype == uris.ui_off) record_on = false;
+
+			// Linked / unlinked to shared data
+			else if (obj2->body.otype == uris.notify_sharedDataLinkEvent)
+			{
+				LV2_Atom *oNr = NULL;
+
+				lv2_atom_object_get
+				(
+					obj2,
+					uris.notify_sharedDataNr, &oNr,
+					NULL
+				);
+
+				if (oNr && (oNr->type == uris.atom_Int))
+				{
+					const int nr = ((LV2_Atom_Int*)oNr)->body;
+
+					if ((nr >= 0) && (nr <= 4) && (nr != sharedDataNr))
+					{
+						if (sharedDataNr != 0) sharedData[sharedDataNr - 1].unlink (this);
+
+						if ((nr != 0) && sharedData[nr - 1].empty())
+						{
+							for (int i = 0; i < NrControllers; ++i) sharedData[nr - 1].set (i, controllers[i]);
+						}
+
+						if (nr != 0) sharedData[nr - 1].link (this);
+						sharedDataNr = nr;
+						notify_sharedData = true;
+
+						for (int i = 0; i < NrControllers; ++i)
+						{
+							float newValue = readController (i);
+							if (newValue != controllers[i])
+							{
+								setController (i, newValue);
+								notify_controllers[i] = true;
+							}
+						}
+					}
+				}
+			}
+
+			// Controller changed
+			else if ((obj2->body.otype == uris.notify_controllerEvent) && (sharedDataNr != 0))
+			{
+				LV2_Atom *oNr = NULL, *oVal = NULL;
+
+				lv2_atom_object_get
+				(
+					obj2,
+					uris.notify_controllerNr, &oNr,
+					uris.notify_controllerValue, &oVal,
+					NULL
+				);
+
+				if (oNr && (oNr->type == uris.atom_Int) && oVal && (oVal->type == uris.atom_Float))
+				{
+					const int nr =  ((LV2_Atom_Int*)oNr)->body;
+
+					if ((nr >= 0) && (nr < NrControllers))
+					{
+						const float val = controllerLimits[nr].validate(((LV2_Atom_Float*)oVal)->body);
+						if ((sharedDataNr > 0) && (sharedDataNr <= 4)) sharedData[sharedDataNr - 1].set (nr, val);
+						setController (nr, val);
+					}
+				}
+			}
 		}
 		ev2 = lv2_atom_sequence_next(ev2);
 	}
@@ -225,7 +277,7 @@ void BChoppr::run (uint32_t n_samples)
 			if (oBbeat && (oBbeat->type == uris.atom_Float))
 			{
 				// Get position within a sequence (0..1)
-				float barsequencepos = ((LV2_Atom_Float*)oBbeat)->body * sequencesperbar / beatsPerBar; // Position within a bar (0..sequencesperbar)
+				float barsequencepos = ((LV2_Atom_Float*)oBbeat)->body * controllers[SequencesPerBar - Controllers] / beatsPerBar; // Position within a bar (0..sequencesperbar)
 				position = MODFL (barsequencepos);			// Position within a sequence
 				refFrame = ev1->time.frames;				// Reference frame
 			}
@@ -237,7 +289,7 @@ void BChoppr::run (uint32_t n_samples)
 	if (last_t < n_samples) play(last_t, n_samples);		// Play remaining samples
 
 	// Update position in case of no new barBeat submitted on next call
-	double relpos = (n_samples - refFrame) * speed / (rate / (bpm / 60)) * sequencesperbar / beatsPerBar;	// Position relative to reference frame
+	double relpos = (n_samples - refFrame) * speed / (rate / (bpm / 60)) * controllers[SequencesPerBar - Controllers] / beatsPerBar;	// Position relative to reference frame
 	position = MODFL (position + relpos);
 	refFrame = 0;
 
@@ -245,6 +297,8 @@ void BChoppr::run (uint32_t n_samples)
 	if (record_on)
 	{
 		notifyGUI();
+		if (notify_sharedData) notifySharedDataNrToGui();
+		for (int i = 0; i < NrControllers; ++i) if (notify_controllers[i]) notifyControllerToGui (i);
 		if (message.isScheduled ()) notifyMessageToGui ();
 	}
 
@@ -252,9 +306,49 @@ void BChoppr::run (uint32_t n_samples)
 	lv2_atom_forge_pop(&forge, &notify_frame);
 }
 
+float BChoppr::readController (const int ctrlNr)
+{
+	// Sync with control ports
+	if ((sharedDataNr == 0) && controlPtrs[ctrlNr]) return controllerLimits[ctrlNr].validate (*controlPtrs[ctrlNr]);
+
+	// Otherwise sync with globally shared data
+	if (sharedDataNr <= 4) return controllerLimits[ctrlNr].validate (sharedData[sharedDataNr - 1].get (ctrlNr));
+
+	return controllerLimits[ctrlNr].min;
+}
+
+void BChoppr::setController (const int ctrlNr, const float value)
+{
+	controllers[ctrlNr] = value;
+
+	if (ctrlNr == Swing - Controllers) recalculateAutoPositions();
+
+	else if (ctrlNr == NrSteps - Controllers) recalculateAutoPositions();
+
+	else if ((ctrlNr >= StepPositions - Controllers) && (ctrlNr < StepPositions - Controllers + MAXSTEPS - 1))
+	{
+		const int step = ctrlNr - (StepPositions - Controllers);
+		if (value == 0.0f)
+		{
+			if (!stepAutoPositions[step])
+			{
+				stepAutoPositions[step] = true;
+				recalculateAutoPositions();
+			}
+		}
+
+		else if (stepPositions[step] != value)
+		{
+			stepAutoPositions[step] = false;
+			stepPositions[step] = value;
+			recalculateAutoPositions();
+		}
+	}
+}
+
 void BChoppr::recalculateAutoPositions ()
 {
-	int nrMarkers = nrSteps - 1;
+	int nrMarkers = controllers[NrSteps - Controllers] - 1;
 	int start = 0;
 	for (int i = 0; i < nrMarkers; ++i)
 	{
@@ -262,6 +356,7 @@ void BChoppr::recalculateAutoPositions ()
 		{
 			if ((i == nrMarkers - 1) || (!stepAutoPositions[i + 1]))
 			{
+				double swing = controllers[Swing - Controllers];
 				double s = 2.0 * swing / (swing + 1.0);
 				double anc = (start == 0 ? 0 : stepPositions[start - 1]);
 				double suc = (i == nrMarkers - 1 ? 1 : stepPositions[i + 1]);
@@ -280,6 +375,32 @@ void BChoppr::recalculateAutoPositions ()
 	}
 }
 
+void BChoppr::notifyControllerToGui (const int nr)
+{
+	// Send notifications
+	LV2_Atom_Forge_Frame frame;
+	lv2_atom_forge_frame_time (&forge, 0);
+	lv2_atom_forge_object (&forge, &frame, 0, uris.notify_controllerEvent);
+	lv2_atom_forge_key (&forge, uris.notify_controllerNr);
+	lv2_atom_forge_int (&forge, nr);
+	lv2_atom_forge_key (&forge, uris.notify_controllerValue);
+	lv2_atom_forge_float (&forge, controllers[nr]);
+	lv2_atom_forge_pop (&forge, &frame);
+	notify_controllers[nr] = false;
+}
+
+void BChoppr::notifySharedDataNrToGui ()
+{
+	// Send notifications
+	LV2_Atom_Forge_Frame frame;
+	lv2_atom_forge_frame_time (&forge, 0);
+	lv2_atom_forge_object (&forge, &frame, 0, uris.notify_sharedDataLinkEvent);
+	lv2_atom_forge_key (&forge, uris.notify_sharedDataNr);
+	lv2_atom_forge_int (&forge, sharedDataNr);
+	lv2_atom_forge_pop (&forge, &frame);
+	notify_sharedData = false;
+}
+
 void BChoppr::notifyGUI()
 {
 	if (record_on)
@@ -344,7 +465,7 @@ void BChoppr::notifyMessageToGui()
 
 void BChoppr::play(uint32_t start, uint32_t end)
 {
-	int steps = nrSteps;
+	int steps = controllers[NrSteps - Controllers];
 
 	//Silence if halted or bpm == 0
 	if ((speed == 0.0f) || (bpm < 1.0f))
@@ -360,10 +481,10 @@ void BChoppr::play(uint32_t start, uint32_t end)
 		float effect2 = audioInput2[i];
 
 		// Interpolate position within the loop
-		float relpos = (i - refFrame) * speed / (rate / (bpm / 60)) * sequencesperbar / beatsPerBar;	// Position relative to reference frame
+		float relpos = (i - refFrame) * speed / (rate / (bpm / 60)) * controllers[SequencesPerBar - Controllers] / beatsPerBar;	// Position relative to reference frame
 		float pos = MODFL (position + relpos);
 
-		if (!bypass)
+		if (!controllers[Bypass - Controllers])
 		{
 			// Calculate step number
 			int iStep;
@@ -384,29 +505,31 @@ void BChoppr::play(uint32_t start, uint32_t end)
 			}
 
 			// Calculate effect (vol) for the position
-			float act = stepLevels[actStep];
-			float prev = stepLevels[prevStep];
-			float next = stepLevels[nextStep];
-			float vol = stepLevels[actStep];
+			float ampSwing = controllers[AmpSwing - Controllers];
+			float* stepLevels = &controllers[StepLevels - Controllers];
+			float act = stepLevels[actStep] * LIM ((actStep % 2 == 0 ? ampSwing : 1.0f / ampSwing), 0, 1);
+			float prev = stepLevels[prevStep] * LIM ((prevStep % 2 == 0 ? ampSwing : 1.0f / ampSwing), 0, 1);
+			float next = stepLevels[nextStep] * LIM ((nextStep % 2 == 0 ? ampSwing : 1.0f / ampSwing), 0, 1);
+			float vol = stepLevels[actStep] * LIM ((actStep % 2 == 0 ? ampSwing : 1.0f / ampSwing), 0, 1);
 
 			// On attack
-			if (iStepFrac < attack)
+			if (iStepFrac < controllers[Attack - Controllers])
 			{
 				if (prev < act)
 				{
-					if (blend == 1) vol = prev + (iStepFrac / attack) * (vol - prev);
-					else if (blend == 2) vol = prev + 0.5 * (sin (M_PI * (iStepFrac / attack - 0.5)) + 1) * (vol - prev);
+					if (controllers[Blend - Controllers] == 1) vol = prev + (iStepFrac / controllers[Attack - Controllers]) * (vol - prev);
+					else if (controllers[Blend - Controllers] == 2) vol = prev + 0.5 * (sin (M_PI * (iStepFrac / controllers[Attack - Controllers] - 0.5)) + 1) * (vol - prev);
 				}
 
 			}
 
 			// On release
-			if (iStepFrac > (1 - release))
+			if (iStepFrac > (1 - controllers[Release - Controllers]))
 			{
 				if (next < act)
 				{
-					if (blend == 1) vol = next + (((1 - iStepFrac)) / release) * (vol - next);
-					else if (blend == 2) vol = next + 0.5 * (sin (M_PI * ((1 - iStepFrac) / release - 0.5)) + 1) * (vol - next);
+					if (controllers[Blend - Controllers] == 1) vol = next + (((1 - iStepFrac)) / controllers[Release - Controllers]) * (vol - next);
+					else if (controllers[Blend - Controllers] == 2) vol = next + 0.5 * (sin (M_PI * ((1 - iStepFrac) / controllers[Release - Controllers] - 0.5)) + 1) * (vol - next);
 				}
 			}
 
@@ -445,11 +568,75 @@ void BChoppr::play(uint32_t start, uint32_t end)
 		}
 
 		// Send effect to audio output
+		const float drywet = controllers[DryWet - Controllers];
 		audioOutput1[i] = audioInput1[i] * (1 - drywet) + effect1 * drywet;
 		audioOutput2[i] = audioInput1[i] * (1 - drywet) + effect2 * drywet;
 	}
 }
 
+LV2_State_Status BChoppr::state_save (LV2_State_Store_Function store, LV2_State_Handle handle, uint32_t flags, const LV2_Feature* const* features)
+{
+	// Save shared data
+	store (handle, uris.notify_sharedDataNr, &sharedDataNr, sizeof(sharedDataNr), uris.atom_Int, LV2_STATE_IS_POD);
+	if (sharedDataNr != 0)
+	{
+		Atom_Controllers atom;
+		for (int i = 0; i < NrControllers; ++i) atom.data[i] = sharedData[sharedDataNr - 1].get (i);
+		atom.body.child_type = uris.atom_Float;
+		atom.body.child_size = sizeof(float);
+		store (handle, uris.notify_controllers, &atom, NrControllers * sizeof (float) + sizeof(LV2_Atom_Vector_Body), uris.atom_Vector, LV2_STATE_IS_POD);
+	}
+
+	return LV2_STATE_SUCCESS;
+}
+
+LV2_State_Status BChoppr::state_restore (LV2_State_Retrieve_Function retrieve, LV2_State_Handle handle, uint32_t flags, const LV2_Feature* const* features)
+{
+	size_t   size;
+	uint32_t type;
+	uint32_t valflags;
+
+	// Restore sharedDataNr
+	if (sharedDataNr != 0) sharedData[sharedDataNr - 1].unlink (this);
+	sharedDataNr = 0;
+	const void* sharedDataNrData = retrieve (handle, uris.notify_sharedDataNr, &size, &type, &valflags);
+	if (sharedDataNrData && (type == uris.atom_Int))
+	{
+		const int nr = *(int*)sharedDataNrData;
+		if (nr != 0) sharedData[nr - 1].link (this);
+		sharedDataNr = nr;
+	}
+
+	// Restore sharedData
+	if ((sharedDataNr >= 0) && (sharedDataNr <= 4))
+	{
+		const void* controllersData = retrieve (handle, uris.notify_controllers, &size, &type, &valflags);
+		if (controllersData && (type == uris.atom_Vector) && (sharedDataNr > 0))
+		{
+			const Atom_Controllers* atom = (const Atom_Controllers*) controllersData;
+			if (atom->body.child_type == uris.atom_Float)
+			{
+				for (int i = 0; i < NrControllers; ++i) sharedData[sharedDataNr - 1].set (i, atom->data[i]);
+			}
+		}
+	}
+
+	notify_sharedData = true;
+
+	// Load controllers
+	for (int i = 0; i < NrControllers; ++i)
+	{
+		float newValue = readController (i);
+		if (newValue != controllers[i])
+		{
+			setController (i, newValue);
+			notify_controllers[i] = true;
+		}
+	}
+
+	return LV2_STATE_SUCCESS;
+}
+
 LV2_Handle instantiate (const LV2_Descriptor* descriptor, double samplerate, const char* bundle_path, const LV2_Feature* const* features)
 {
 	// New instance
@@ -489,12 +676,37 @@ void run (LV2_Handle instance, uint32_t n_samples)
 	inst->run (n_samples);
 }
 
+static LV2_State_Status state_save(LV2_Handle instance, LV2_State_Store_Function store, LV2_State_Handle handle, uint32_t flags,
+           const LV2_Feature* const* features)
+{
+	BChoppr* inst = (BChoppr*)instance;
+	if (!inst) return LV2_STATE_SUCCESS;
+
+	return inst->state_save (store, handle, flags, features);
+}
+
+static LV2_State_Status state_restore(LV2_Handle instance, LV2_State_Retrieve_Function retrieve, LV2_State_Handle handle, uint32_t flags,
+           const LV2_Feature* const* features)
+{
+	BChoppr* inst = (BChoppr*)instance;
+	if (!inst) return LV2_STATE_SUCCESS;
+
+	return inst->state_restore (retrieve, handle, flags, features);
+}
+
 void cleanup (LV2_Handle instance)
 {
 	BChoppr* inst = (BChoppr*) instance;
 	delete inst;
 }
 
+static const void* extension_data(const char* uri)
+{
+	static const LV2_State_Interface state  = {state_save, state_restore};
+	if (!strcmp(uri, LV2_STATE__interface)) return &state;
+	return NULL;
+}
+
 const LV2_Descriptor descriptor =
 {
 		BCHOPPR_URI,
@@ -504,7 +716,7 @@ const LV2_Descriptor descriptor =
 		run,
 		NULL, //deactivate,
 		cleanup,
-		NULL //extension_data
+		extension_data
 };
 
 // LV2 Symbol Export
